// Turret servo controller
// Written by Yvo de Haas

#include <Servo.h> 

Servo gunpan, guntilt, eyepan, eyetilt;

const int eyeled = 3;
const int eyelaser = 2;
const int gun1 = 12;
const int gun2 = 13;
const int trigger = A0;

const int joystickpanpin = 1;
const int joysticktiltpin = 2;

boolean Serialcontrol = 0;

int joystickpan, joysticktilt;
int gunpantarget, gunpancurrent, guntilttarget, guntiltcurrent;
int eyepantarget, eyepancurrent, eyetilttarget, eyetiltcurrent;
int jitterpanoffset, jittertiltoffset;
int readbuffer[10];
byte readbufferpos = 0;
int xpos,ypos,gunpos;
int temp_compensation;
unsigned long serial_start_time;
boolean guntoggle = 0;
unsigned long millisdelaytimer;
unsigned long millisjittertimer;
unsigned long millisguntimer;

const int eyepanmin = 45;
const int eyepanmax = 135;
const int eyepaninverted = 1;
const int eyepanspeed = 1; //in degrees per cycle (default 100ms)
const int serialeyepaninverted = 1;
const int eyetiltmin = 70;
const int eyetiltmax = 120;
const int eyetiltinverted = 0;
const int eyetiltspeed = 1; //in degrees per cycle (default 100ms)
const int serialeyetiltinverted = 0;


const int gunpanmin = 20;
const int gunpanmax = 180;
const int gunpaninverted = 1;
const int gunpanspeed = 1; //in degrees per cycle (default 100ms)
const int guntiltmin = 60;
const int guntiltmax = 110;
const int guntiltinverted = 1;
const int guntiltspeed = 1; //in degrees per cycle (default 100ms)

const int gunpantilt_compensation = 2; //in degrees of pan per degree of tilt

const int jitteron = 1;

const int camerax = 640;
const int cameray = 480;

const int millisdelaytime = 20; //in milliseconds (default 100ms)
const int jitterdelaytime = 150; //in milliseconds
const int gunflashdelay = 40; //in milliseconds

const long serial_return_time = 5000; //in milliseconds, no serial time before the joystick takes over again



void setup()
{
  Serial.begin(9600); 

  pinMode(eyeled, OUTPUT);
  pinMode(eyelaser, OUTPUT);
  pinMode(gun1, OUTPUT);
  pinMode(gun2, OUTPUT);
  pinMode(trigger, INPUT);

  digitalWrite(eyeled, 1);
  //delay(1500);
  digitalWrite(eyelaser, 1);

  eyepan.attach(5);
  eyetilt.attach(6);
  gunpan.attach(9);
  guntilt.attach(10);

  //set tilt and pan values to center
  eyepancurrent = map(1, 0, 2, eyepanmin, eyepanmax);
  eyetiltcurrent = map(1, 0, 2, eyetiltmin, eyetiltmax);
  gunpancurrent = map(1, 0, 2, gunpanmin, gunpanmax);
  guntiltcurrent = map(1, 0, 2, guntiltmin, guntiltmax);

  millisdelaytimer = millis();
}

void loop()
{
  if (millis() > millisdelaytimer) //non interruptive delay
  {
    millisdelaytimer += millisdelaytime;

    if(Serialcontrol == 0)
    {
      if (Serial.available() > 0)//turn on serial control when serial line is found
      {
        Serialcontrol = 1;
      }

      //if no serial line is available, move using joystick
      //measure joystick values
      joystickpan = analogRead(joystickpanpin);
      joysticktilt = analogRead(joysticktiltpin);

      //remap values
      //invert where necessary
      if(eyepaninverted == 1)
      {
        eyepantarget = map(joystickpan, 0, 1023, eyepanmin, eyepanmax);
      }
      else
      {
        eyepantarget = map(joystickpan, 1023, 0, eyepanmin, eyepanmax);
      }
      if(eyetiltinverted == 1)
      {
        eyetilttarget = map(joysticktilt, 0, 1023, eyetiltmin, eyetiltmax);
      }
      else
      {
        eyetilttarget = map(joysticktilt, 1023, 0, eyetiltmin, eyetiltmax);
      }
    }
    else
    {
      if (Serial.available() > 0)
      {
        //serial_start_time = millis();
        readbuffer[readbufferpos] = Serial.read();
        if (readbuffer[readbufferpos] == 84 || readbuffer[readbufferpos] == 88 || readbuffer[readbufferpos] == 89)
        {
          if (readbuffer[readbufferpos] == 88)
          {
            readbufferpos ++;
            ypos = Serial.parseInt();
          }
          else if (readbuffer[readbufferpos] == 89)
          {
            readbufferpos ++;
            xpos = Serial.parseInt();
          }
          else if (readbuffer[readbufferpos] == 84)
          {
            readbufferpos ++;
            gunpos = Serial.parseInt();
          }
        }
        else if (readbuffer[readbufferpos] == 13 || readbuffer[readbufferpos] == 32)
        {
          exportbuffer();
          readbufferpos = 0;
        }
        else if (readbuffer[readbufferpos] == 10)
        {
        }
        else
        {
          readbufferpos ++;
        }     
      }      
      if(serialeyepaninverted == 1)
      {
        eyepantarget = map(xpos, 0, cameray, eyepanmin, eyepanmax);
      }
      else
      {
        eyepantarget = map(xpos, cameray, 0, eyepanmin, eyepanmax);
      }
      if(serialeyetiltinverted == 1)
      {
        eyetilttarget = map(ypos, 0, camerax, eyetiltmin, eyetiltmax);
      }
      else
      {
        eyetilttarget = map(ypos, camerax, 0, eyetiltmin, eyetiltmax);
      }
    }


    if (digitalRead(trigger) == 0 || gunpos == 1) //trigger offset
    { //trigger recalculation (adds random jitter to all axes)
      if (jitteron == 1)
      {
        if (millis() > millisjittertimer )
        {
          millisjittertimer = millis() + jitterdelaytime;
          jitterpanoffset = random(8);
          jittertiltoffset = random(8);
        }

        //add jittering offsets
        eyepantarget -= 4;
        eyepantarget += jitterpanoffset;
        gunpantarget -= 4;
        gunpantarget += jitterpanoffset;
        eyetilttarget -= 4;
        eyetilttarget += jittertiltoffset;
        guntilttarget -= 4;
        guntilttarget += jittertiltoffset;
      }

      if (millis() > millisguntimer)
      {
        millisguntimer = millis() + gunflashdelay;
        if (guntoggle == 1)
        {
          guntoggle = 0;
          digitalWrite(gun1, 1);
          digitalWrite(gun2, 0);
        }
        else
        {
          guntoggle = 1;
          digitalWrite(gun1, 0);
          digitalWrite(gun2, 1);
        }
      }
    }
    else
    {
      digitalWrite(gun1, 0);
      digitalWrite(gun2, 0);
      guntoggle = 0;
    }

    //recalculate new current positions based on speed
    //see if trigger is on, if so recalculate additional position based on gun movement
    if (eyepantarget > eyepancurrent)
    {
      eyepancurrent += eyepanspeed;
      if (eyepancurrent > eyepantarget) //overflow protection
      {
        eyepancurrent == eyepantarget;
      }
    }
    else if (eyepantarget < eyepancurrent)
    {
      eyepancurrent -= eyepanspeed;
      if (eyepancurrent < eyepantarget) //overflow protection
      {
        eyepancurrent == eyepantarget;
      }
    }
    if (eyetilttarget > eyetiltcurrent)
    {
      eyetiltcurrent += eyetiltspeed;
      if (eyetiltcurrent > eyetilttarget) //overflow protection
      {
        eyetiltcurrent == eyetilttarget;
      }
    }
    else if (eyetilttarget < eyetiltcurrent)
    {
      eyetiltcurrent -= eyetiltspeed;
      if (eyetiltcurrent < eyetilttarget) //overflow protection
      {
        eyetiltcurrent == eyetilttarget;
      }
    }

    if(gunpaninverted == 1)
    {
      gunpancurrent = map(eyepancurrent, eyepanmin, eyepanmax, gunpanmin, gunpanmax);
    }
    else
    {
      gunpancurrent = map(eyepancurrent, eyepanmax, eyepanmin, gunpanmin, gunpanmax);
    }
    if(guntiltinverted == 1)
    {
      guntiltcurrent = map(eyetiltcurrent, eyetiltmin, eyetiltmax, guntiltmin, guntiltmax);
    }
    else
    {
      guntiltcurrent = map(eyetiltcurrent, eyetiltmax, eyetiltmin, guntiltmin, guntiltmax);
    }

    //gun pan tilt compensation
    temp_compensation = guntiltmin + guntiltmax;
    temp_compensation /= 2;
    temp_compensation = guntiltcurrent - temp_compensation;
    temp_compensation /= gunpantilt_compensation;
    gunpancurrent -= temp_compensation;

    //write new servo values
    eyepan.write(eyepancurrent);
    eyetilt.write(eyetiltcurrent);
    gunpan.write(gunpancurrent);
    guntilt.write(guntiltcurrent);

    //break from serial control
    /*if (Serialcontrol == 1 && serial_start_time + serial_return_time < millis())
     {
     Serialcontrol = 0;
     }*/
  }
}



void exportbuffer()
{
  for (byte b = 0; b < 6; b++)
  {
    readbuffer[b] = 0;
  }    
}









